/*
 * Copyright (c) <2008>, <Hiroshi Sakakibara >
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the
 * distribution.  Neither the name of the <ORGANIZATION> nor the names
 * of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written
 * permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */
/*
 * X3F file extractor 
 * 
 * 1 Sep 2008
 * - first 
 * skk@ht.sfc.keio.ac.jp 
 */

#include <iconv.h>

#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>


#include <sys/types.h>
#include <sys/stat.h>

//#define dbg(...) printf("#DEBUG: "__VA_ARGS__)
#define dbg

struct version {
	unsigned short low; 
	unsigned short high; 
}; 

struct prop_header {
	unsigned char sec_id[4];  // should be "SECp"	
	struct version v; 
	unsigned int entry_num; 
	unsigned int char_format; 
	unsigned int reserved; 
	unsigned int length; 
}; 

struct ima2_header {
	unsigned char sec_id[4];  // should be "SECi"
	struct version v; 
	unsigned int type; 
	unsigned int data_format; // 18=jpeg
	unsigned int columns; 
	unsigned int rows; 
	unsigned int row_size; 
}; 

struct dir_entry {
	unsigned int start_offset; 
	unsigned int length; 
	unsigned char type[4]; 
}; 

struct dir_sec {
	unsigned char section_identifier[4]; 
	struct version v; 
	unsigned int dir_entry_num; 
}; 

struct dir_pointer {
	unsigned int pointer; 
}; 

// X3F header. 
// valid for version 2.1+ 
struct header {
	unsigned char filetype[4]; 
	struct version v; 
	unsigned char unique_identifier[16]; 
	unsigned int mark_bits; 
	unsigned int image_columns; 
	unsigned int image_rows; 
	unsigned int rotation; 
	unsigned char white_balance_label[32]; 
	unsigned char extended_data_type[32]; 
	float extended_data[32]; 
}; 

int read_x3f(char *); 
int read_prop(char *, unsigned int, unsigned int); 
int read_ima2(char *, unsigned int, unsigned int); 
void usage(void); 
char * utf2ascii(char *, int); 
int list_note(char *, char *); 
int output_jpeg(char *, unsigned int, unsigned int); 

char filename[BUFSIZ]; 
int stdoutput; 
int verbose; 

int
main(argc, argv)
	int argc; 
	char **argv; 
{
	int c; 
        extern char *optarg;


	if (argc < 2) {
		usage(); 
		exit(1); 
	}

	stdoutput = 0; 
	verbose = 0; 
        while((c = getopt(argc, argv, "cvh")) != -1){
                switch(c){
                case 'c':
			stdoutput = 1; 
                        break;
                case 'v':
			verbose = 1; 
                        break;
                case 'h':
                        usage();
                        exit(0);
                default:
                        usage();
                        break;
                }
        }

	memset(filename, 0, sizeof(filename)); 
	strcpy(filename, argv[argc-1]); 

	if (read_x3f(filename) < 0) { 
		exit(1); 
	}

	exit(0); 
}


int
read_x3f(filename)
	char *filename; 
{
	int fd; 
	int i, j; 
	int biggest_size; 
	int biggest_entry_num ;
	char *buf; 
	struct stat sb; 
	struct header *header_p; 
	struct dir_pointer *dp; 
	struct dir_sec *dir_p; 
	struct dir_entry *dir_entry_p; 

	memset(&sb, 0, sizeof(sb)); 
	if (stat(filename, &sb) < 0) { 
		perror("stat"); 
		return(-1);
	}

	buf = (char *)malloc(sb.st_size); 
	memset(buf, 0, sb.st_size); 
	if (buf == NULL) {
		perror("malloc"); 
		return(-1); 
	}

	fd = open(filename, O_RDONLY); 
	if (fd < 0) {
		perror("open"); 
		return(-1); 
	}


	if (read(fd, buf, sb.st_size) < 0) {
		perror("read"); 
		return(-1); 
	}
	close(fd); 


	/* header */
	header_p = (struct header *)buf; 
	if (verbose) {
		char tmp[20]; 

		printf("X3F Header\n"); 
		memset(tmp, 0, sizeof(tmp)); 
		memcpy(tmp, header_p->filetype, 4); 
		printf("\tFile type identifier: %s\n", tmp); 
		printf("\tFile format version: %d.%d\n", header_p->v.high, header_p->v.low); 
		printf("\tUnique identifier: "); 
		for (j=0; j<sizeof(header_p->unique_identifier); j++) {
			printf("%x", header_p->unique_identifier[j]); 
		}
		printf("\n"); 
		printf("\tMark bits: %d\n", header_p->mark_bits); 
		printf("\tImage columns: %d\n", header_p->image_columns); 
		printf("\tImage rows: %d\n", header_p->image_rows); 
		printf("\tRotation: %d\n", header_p->rotation); 
		printf("\tWhite balance label string: %s\n", header_p->white_balance_label); 
		printf("\tExtended_data_types: \n"); 
		for (j=0; j<32; j++) {
			switch(header_p->extended_data_type[j]){
			case 0: 
				printf("\t\tslot %d: unused\n", j); 
				break; 
			case 1: 
				printf("\t\tslot %d: Exposure adjust:\t%f\n", j, header_p->extended_data[j]); 
				break; 
			case 2: 
				printf("\t\tslot %d: Contarst adjust:\t%f\n", j, header_p->extended_data[j]); 
				break; 
			case 3: 
				printf("\t\tslot %d: Shadow adjust:\t\t%f\n", j, header_p->extended_data[j]); 
				break; 
			case 4: 
				printf("\t\tslot %d: Highlight adjust:\t%f\n", j, header_p->extended_data[j]); 
				break; 
			case 5: 
				printf("\t\tslot %d: Saturation adjust:\t%f\n", j, header_p->extended_data[j]); 
				break; 
			case 6: 
				printf("\t\tslot %d: Sharpness adjust:\t%f\n", j, header_p->extended_data[j]);
				break; 
			case 7: 
				printf("\t\tslot %d: Color adjust red:\t%f\n", j, header_p->extended_data[j]);
				break; 
			case 8: 
				printf("\t\tslot %d: Color adjust green:\t%f\n", j, header_p->extended_data[j]);
				break; 
			case 9: 
				printf("\t\tslot %d: Color adjust blue:\t%f\n", j, header_p->extended_data[j]);
				break; 
			case 10: 
				printf("\t\tslot %d: X3 Fill Light adjust:\t%f\n", j, header_p->extended_data[j]);
				break; 
			default: 
				printf("\t\tslot %d: invalid data type\n", j); 
			}
		}
	}

	/* directory pointer */
	dp = (struct dir_pointer *)(buf+sb.st_size-4); 

	/* directory */
	dir_p = (struct dir_sec *)(buf+dp->pointer); 
	if (verbose) {
		printf("Directory Section\n"); 
		printf("\tSection identifier: %s\n", dir_p->section_identifier); 
		printf("\tSection version: %x.%x\n", dir_p->v.high, dir_p->v.low); 
		printf("\tNumber of directory entries: %d\n", dir_p->dir_entry_num); 
	}

	/* get biggest jpeg entry number */
	biggest_size = 0;       // init 
	biggest_entry_num = -1; // init
	for (i=0; i<dir_p->dir_entry_num; i++) {
		dir_entry_p = (struct dir_entry *)(((struct dir_entry *)(dir_p + 1)) + i);

		if (memcmp(dir_entry_p->type, "IMA2", 4) == 0) {
			char *image; 
			struct ima2_header *ih; 

			image = buf + dir_entry_p->start_offset; 
			ih = (struct ima2_header *)image; 
			
			if (ih->type == 2) { // processed for preview 
				if (dir_entry_p->length > biggest_size) {
					biggest_size = dir_entry_p->length; 
					biggest_entry_num = i; 
				}
			}
		}
	}

	for (i=0; i<dir_p->dir_entry_num; i++) {
		dir_entry_p = (struct dir_entry *)(((struct dir_entry *)(dir_p + 1)) + i);

		if (memcmp(dir_entry_p->type, "PROP", 4) == 0) {
			if (verbose) {
				printf("\tDirectory entry %d, Property List(PROP)\n", i); 
				read_prop(buf, dir_entry_p->start_offset, dir_entry_p->length); 
			}
		} else if (memcmp(dir_entry_p->type, "IMAG", 4) == 0) {
			; 
			//dbg("IMAG\n"); 
		} else if (memcmp(dir_entry_p->type, "IMA2", 4) == 0) {
			if (verbose) {
				printf("\tDirectory entry %d, Image data(IMA2)\n", i); 
			}
			read_ima2(buf, dir_entry_p->start_offset, dir_entry_p->length); 
		} else {
			if (verbose) {
				char tmp[5]; 
				memset(tmp, 0, sizeof(tmp)); 
				memcpy(tmp, dir_entry_p->type, 4); 
				printf("\tDirectory entry %d, Unknown(%s)\n", i, tmp); 
			}
		}

		if (verbose == 0 && i == biggest_entry_num) {
			output_jpeg(buf, dir_entry_p->start_offset, dir_entry_p->length); 
		}
	}

	free(buf); 

	return(0); 
}

int
output_jpeg(buf, offset, length)
	char *buf; 
	unsigned int offset; 
	unsigned int length; 
{
	int fd; 
	char *image; 
	char cwd[BUFSIZ]; 
	char out_name[BUFSIZ]; 
	struct ima2_header *ih; 

	image = buf+offset; 
	ih = (struct ima2_header *)image; 

	/* copies  the jpeg file to the file or stdout */

	if (stdoutput == 0) {
		getcwd(cwd, sizeof(cwd)); 

		strcpy(out_name, cwd); 
		strcat(out_name, "/"); 
		strcat(out_name, filename); 
		strcat(out_name, ".jpg"); 

		fd = open(out_name, O_CREAT|O_WRONLY); 
		if (fd < 0) {
			perror("open"); 
			exit(1); 
		}
	} else if (stdoutput == 1) {
		fd = 1; 
	}

	if (write(fd, image+sizeof(struct ima2_header), length-sizeof(struct ima2_header)) < 0) {
		perror("write"); 
		close(fd); 
		exit(1); 
	}
	if (stdoutput == 0) {
		close(fd); 
		chmod (out_name, S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH); 
	}
	
	return(0);
}

int
read_prop(buf, offset, length)
	char *buf; 
	unsigned int offset; 
	unsigned int length; 
{
	int i; 
	struct prop_header *ph; 
	struct namevalue {
		unsigned int name_offset; 
		unsigned int value_offset; 
	}; 
	struct namevalue *nv; 

	if (verbose) {
		ph = (struct prop_header *)(buf + offset); 

		printf("\t\tSection identifier: %s\n", ph->sec_id); 
		printf("\t\tProperty list format version: %d.%d\n", ph->v.high, ph->v.low); 
		printf("\t\tNumber of property entries: %d\n", ph->entry_num); 
		printf("\t\tCharacter format for all entries in this table: %d\n", ph->char_format); 
		printf("\t\t (0:  CHAR16 Unicode)\n"); 
		printf("\t\tTotal length of name/value data in characters: %d\n", ph->length); 

		//data = buf + offset + (sizeof(struct prop_header)) + ph->entry_num*8; 

		printf("\t\tProperty List: \n"); 
		nv = (struct namevalue *)(buf + offset + sizeof(struct prop_header)); 

		for (i=0; i<ph->entry_num; i++, nv++) {
			int value_length; 
			char *name_ascii; 
			char *value_ascii; 
			char *char_data_start; 
			char_data_start = ((char *)(ph+1)) + ph->entry_num * 8; 

			name_ascii = utf2ascii(char_data_start + nv->name_offset * 2,
					       nv->value_offset*2 - nv->name_offset*2); 
			if (i != ph->entry_num - 1) {
				value_length = (nv+1)->name_offset*2 - nv->value_offset*2; 
			} else {
				value_length = (ph->length - nv->value_offset) * 2; 
			}
			value_ascii = utf2ascii(char_data_start + nv->value_offset * 2,
						value_length); 
			printf("\t\t\tList %d, %s =\t%s\n", 
			       i, 
			       name_ascii, 
			       value_ascii); 
			list_note(name_ascii, value_ascii); 


			free(name_ascii); 
			free(value_ascii); 

		}
	}
	return(0); 
}

int
list_note(name, value)
	char *name;
	char *value; 
{
	if (strcmp(name, "AEMODE") == 0) {
		printf("\t\t\t (8 = 8-segment, C = center-weighted, A = average-weighted)\n"); 
	} else if (strcmp(name, "AFMODE") == 0) {
		printf("\t\t\t (AF-S = single, AF-C = continuous, MF = manual focus)\n"); 
	} else if (strcmp(name, "AP_DESC") == 0) {
		printf("\t\t\t (Aperture shown on LCD at time of shot)\n"); 
	} else if (strcmp(name, "APERTURE") == 0) {
		printf("\t\t\t (Exact aperture value)\n"); 
	} 
	/* need to copy texts from X3F specification. 
	 * http://www.photofo.com/downloads/x3f-raw-format.pdf
	 */ 

	return(0); 
}

char * 
utf2ascii(inbuf, inbuflen)
	char *inbuf; 
	int inbuflen; 
{
        iconv_t ic;
        char *ret;
        char *asciibuf_p;
        char *utfbuf_p;

        char asciibuf[BUFSIZ+1];
        char utfbuf[BUFSIZ+1];

        int num;
        size_t asciibuflen = BUFSIZ;
        size_t utfbuflen = BUFSIZ;

//	dbg("inbuflen: %d, %d\n", inbuflen, strlen(inbuf)); 

        memset(asciibuf, 0, BUFSIZ+1);
        memset(utfbuf, 0, BUFSIZ+1);

	memcpy(utfbuf, inbuf, inbuflen); 

        asciibuf_p = asciibuf;
        utfbuf_p = utfbuf;

        ic = iconv_open("ASCII", "UCS-2LE");
        num = iconv(ic, &utfbuf_p, &utfbuflen, &asciibuf_p, &asciibuflen);
        if (num < 0){
                perror("iconv:");
        }
        iconv_close(ic);

        ret = (char *)malloc(BUFSIZ);
        strcpy(ret, asciibuf);

        return(ret);

}


int
read_ima2(buf, offset, length)
	char *buf; 
	unsigned int offset; 
	unsigned int length; 
{
	char *image; 
	struct ima2_header *ih; 

	image = buf+offset; 
	ih = (struct ima2_header *)image; 

	if (verbose) { 
		printf("\t\tSection identifier: %s\n", ih->sec_id); 
		printf("\t\tImage format version: %d.%d\n", ih->v.high, ih->v.low);
		printf("\t\tType of image data: %d\n", ih->type); 
		printf("\t\t (2: processed for preview)\n"); 
		printf("\t\tData format: %d\n", ih->data_format); 
		printf("\t\t (3:  uncompressed 24-bit 8/8/8 RGB,\n"); 
		printf("\t\t (11: Huffman-encoded DPCM 8/8/8 RGB,\n"); 
		printf("\t\t (18: JPEG-compressed 8/8/8 RGB\n"); 
		printf("\t\tImage columns: %d\n", ih->columns); 
		printf("\t\tImage rows: %d\n", ih->rows); 
		printf("\t\tRow size in bytes: %d\n", ih->row_size); 
	}

	return(0); 
}

void
usage()
{
	printf("usage: x3f [-cvh] filename\n"); 
	printf("\t-c: outputs jpeg data to stdout\n"); 
	printf("\t-v: outputs meta information (do not output jpeg data)\n"); 
	printf("\t-h: outputs usage(this message)\n"); 

}
